@@ -1,20 +1,21 @@
 package CORBA::Fixed;
-use Math::BigInt;
 
 use overload 
-     '+' => \&add,
-     '-' => \&subtract,
-     '*' => \&mul,
-     '/' => \&div,
-     '<=>' => \&compare,
-     '""' => \&stringify;
+    '+' => \&add,
+    '-' => \&subtract,
+    '*' => \&mul,
+    '/' => \&div,
+    '<=>' => \&compare,
+    '""' => \&stringify;
+
+require Math::BigInt;
 
 sub _construct {
     my ($class, $value, $scale) = @_;
 
     bless {
-	    v  => $value,
-	   's' => $scale,
+	   v => $value,
+	   s => $scale,
 	  }, $class;
 }
 
@@ -24,7 +25,7 @@ sub from_string {
     my ($leading,$rest) = $str =~ /^(\s*[+-]?\d+)(?:\.(\d+)*)?/;
 
     if (!defined $leading) {
-	return CORBA::Fixed->_construct(new Math::BigInt("0"),0);
+	return CORBA::Fixed->_construct(0,0);
     } else {
 	$rest = defined $rest ? $rest : "";
         $str = $leading.$rest;
@@ -43,6 +44,7 @@ sub from_string {
 
 sub new {
     my ($class, $v, $scale) = @_;
+
     CORBA::Fixed->_construct (Math::BigInt->new($v), $scale);
 }
 
@@ -55,14 +57,12 @@ sub add {
 
     my ($v, $s);
     
-    if ($a->{'s'} > $b->{'s'}) {
-	$s = $a->{'s'};
-        my $z = $b->{v}->copy;
-	$v = $a->{v} + ($z->blsft($a->{'s'} - $b->{'s'},10));
+    if ($a->{s} > $b->{s}) {
+	$s = $a->{s};
+	$v = $a->{v} + ($b->{v}.("0" x ($a->{s} - $b->{s})));
     } else {
-	$s = $b->{'s'};
-        my $z = $a->{v}->copy;
-	$v = $b->{v} + ($z->blsft($b->{'s'} - $a->{'s'},10));
+	$s = $b->{s};
+	$v = $b->{v} + ($a->{v}.("0" x ($b->{s} - $a->{s})));
     }
 
     CORBA::Fixed->_construct ($v, $s);
@@ -81,14 +81,16 @@ sub subtract {
     
     my ($v, $s);
 
-    if ($a->{'s'} > $b->{'s'}) {
-        $s = $a->{'s'};
-        my $z = $b->{v}->copy;
-        $v = $a->{v} - ($z->blsft($a->{'s'} - $b->{'s'},10));
-    } else {
-        $s = $b->{'s'};
-        my $z = $a->{v}->copy;
-        $v = ($z->blsft($b->{'s'} - $a->{'s'},10)) - $b->{v};
+    {
+	local $^W = 0;		# BigInt.pm problems
+
+	if ($a->{s} > $b->{s}) {
+	    $s = $a->{s};
+	    $v = $a->{v} - ($b->{v}.("0" x ($a->{s} - $b->{s})));
+	} else {
+	    $s = $b->{s};
+	    $v = ($a->{v}.("0" x ($b->{s} - $a->{s}))) - $b->{v};
+	}
     }
     CORBA::Fixed->_construct ($v, $s);
 }
@@ -104,14 +106,10 @@ sub compare {
 	($a, $b) = ($b, $a);
     }
     
-    if ($a->{'s'} > $b->{'s'}) {
-        my $z = $b->{v}->copy;
-        $z->blsft($a->{'s'} - $b->{'s'},10);
-	$a->{v} <=> $z;
+    if ($a->{s} > $b->{s}) {
+	$a->{v} <=> ($b->{v}.("0" x ($a->{s} - $b->{s})));
     } else {
-        my $z = $a->{v}->copy;
-        $z->blsft($b->{'s'} - $a->{'s'},10);
-	$z <=> $b->{v};
+	($a->{v}.("0" x ($b->{s} - $a->{s}))) <=> $b->{v};
     }
 }
 
@@ -122,7 +120,7 @@ sub mul {
 	$b = CORBA::Fixed->from_string($b);
     }
 
-    CORBA::Fixed->_construct ($a->{v}*$b->{v}, $a->{'s'}+$b->{'s'});
+    CORBA::Fixed->_construct ($a->{v}*$b->{v}, $a->{s}+$b->{s});
 }
 
 sub div {
@@ -132,21 +130,28 @@ sub div {
 	$b = CORBA::Fixed->from_string($b);
     }
 
+    if ($reverse) {
+	($a, $b) = ($b, $a);
+    }
+    
     # calculate to 31 places
 
-    my $s = ($a->{'s'} - $b->{'s'});
+    my $s = ($a->{s} - $b->{s});
 
     my $v1 = $a->{v};
     my $v2 = $b->{v};
 
-    my $pad = 31 - $v1->length - $v2->length;
+    my $pad = (31 - (length($v1) - length($v2)));
 
     if ($pad > 0) {
-        $v1->blsft( $pad, 10 );
+	$v1 = new Math::BigInt ($v1.("0" x $pad));
 	$s += $pad;
     }
 
-    CORBA::Fixed->_construct ($v1/$v2, $s);
+    {
+	local $^W = 0;		# BigInt.pm problems
+        CORBA::Fixed->_construct ($v1/$v2, $s);
+    }
 }
 
 # Turn the number into a form suitable for turning into a 
@@ -155,13 +160,28 @@ sub div {
 sub to_digits {
     my ($self, $ndigits, $scale) = @_;
 
-    my $value = $self->{v}->copy;
-    if( $self->{'s'} <= $scale ) {
-        $value->blsft( $scale - $self->{'s'}, 10 );
+    my $value = $self->{v};
+    my $vstr = "$value";
+    
+    if ($self->{s} > $scale) {
+	my $rest = substr($vstr, -($self->{s} - $scale));
+	substr($vstr, -($self->{s} - $scale)) = "";
+
+	# Banker's rounding
+	if (length ($rest) > 0) {
+	    my $half = new Math::BigInt ("5".('0' x (length ($rest)-1)));
+	    $rest = new Math::BigInt ($rest);
+	    $value = new Math::BigInt ($vstr);
+
+	    if ($rest == $half) {
+		$vstr = "" . new Math::BigInt ($value + ((substr($vstr,-1) % 2) ? 1 : 0));
+	    } else {
+		$vstr = "" . new Math::BigInt ($value + (($rest < $half) ? 0 : 1));
+	    }
+	}
     } else {
-        $value->brsft( $self->{'s'} - $scale, 10 );
+	$vstr .= '0' x ($scale - $self->{s});
     }
-    my $vstr = (($value->sign eq '+') ? '+' : '') . $value->bstr();
 
     # pad or truncate to the requested number of digits
     my $len = length ($vstr) - 1;
@@ -170,19 +190,21 @@ sub to_digits {
     } else {
        return substr($vstr,0,1) . substr($vstr,-$ndigits);
     }
+
 }
 
 sub stringify {
     my $self = shift;
 
-    my $vstr = $self->{v}->bstr();
-    my $scale = $self->{'s'};
+    my $vstr = "$self->{v}";
+    my $scale = $self->{s};
 
     if ($scale > 0) {
        return substr($vstr,0,length($vstr)-$scale).".".substr($vstr,-$scale);
     } else {
        return $vstr . ('0' x -$scale);
     }
+
 }
 
 1;
